#!/usr/bin/env python3

from math import sqrt, pi, cos, asin, acos
import sys

def length(v):
    return sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2])

def cross(a, b):
    u = a[1]*b[2] - a[2]*b[1]
    v = a[2]*b[0] - a[0]*b[2]
    w = a[0]*b[1] - a[1]*b[0]
    return u,v,w

def norm(v):
    l = length(v)
    return v[0]/l, v[1]/l, v[2]/l


def normalize(x, y, z):
    l = length((x, y, z))
    return (x / l, y / l, z / l)


def get_middle_point(p1, p2):
    return normalize((p1[0] + p2[0]) / 2, (p1[1] + p2[1]) / 2, (p1[2] + p2[2]) / 2)


def jetmap(val, minval, maxval):
    t = (val - minval) / (maxval - minval)
    d = [0,0,0]
    if t <= 0.125:
        d[0] = d[1] = 0.0
        d[2] = 0.5 + 0.5 * (t / 0.125)
    elif t < 0.375:
        d[0] = 0.0
        d[1] = (t - 0.125) / 0.25
        d[2] = 1.0
    elif t < 0.625:
        d[0] = (t - 0.375) / 0.25
        d[1] = 1.0
        d[2] = 1.0 - d[0]
    elif t < 0.875:
        d[0] = 1.0
        d[1] = 1.0 - (t - 0.625) / 0.25
        d[2] = 0.0
    else:
        d[1] = d[2] = 0.0
        d[0] = 1.0 - 0.5 * ((t - 0.875) / 0.125)
    return (d[0]*255, d[1]*255, d[2]*255)


# calculate the circumcircle of a triangle in 3D and return the unit vector of
# the center of the circumcircle as well as the angle theta of the sphere cap
# that has the circumcirle as its base
def circumcircle(v1, v2, v3):
    a = v1[0] - v3[0], v1[1] - v3[1], v1[2] - v3[2]
    b = v2[0] - v3[0], v2[1] - v3[1], v2[2] - v3[2]
    c = a[0] - b[0], a[1] - b[1], a[2] - b[2]
    la = length(a)
    la2 = la ** 2
    lb = length(b)
    lb2 = lb ** 2
    axb = cross(a, b)
    laxb = length(axb)
    laxb2 = 2 * laxb ** 2
    r = (la * lb * length(c)) / (2 * laxb)
    # Now we have the radius of the circumcircle but in the end, instead of the
    # radius, we want half the angle under which the circumcircle is seen from
    # the center of the sphere.
    # In other words: instead of the radius of the base of the spherical cap,
    # we want the angle theta
    theta = asin(r)
    sa = lb2 * a[0], lb2 * a[1], lb2 * a[2]
    sb = la2 * b[0], la2 * b[1], la2 * b[2]
    di = sb[0] - sa[0], sb[1] - sa[1], sb[2] - sa[2]
    p = cross(di, axb)
    p = p[0] / laxb2 + v3[0], p[1] / laxb2 + v3[1], p[2] / laxb2 + v3[2]
    # the point p we are interested in is not the center of the base of the
    # circle cap but the point on the unit-sphere so that we can easily find
    # the angle between it and other vectors later
    p = norm(p)
    return p, theta


def main():
    if len(sys.argv) <= 1:
        print("usage: %s oct|ico depth tri|cc|rat [min] [max]"%sys.argv[0])
        exit(1)

    if sys.argv[1] == "oct":
        points = [
                (-1,0,0),
                (1,0,0),
                (0,-1,0),
                (0,1,0),
                (0,0,-1),
                (0,0,1)
                ]
        faces = []
        for x in [-1,1]:
            for y in [-1,1]:
                for z in [-1,1]:
                    v1 = 0 if x < 0 else 1
                    v2 = 2 if y < 0 else 3
                    v3 = 4 if z < 0 else 5
                    if (x > 0) ^ (y > 0) ^ (z > 0) == False:
                        v1, v3 = v3, v1
                    faces.append((points[v1],points[v2],points[v3]))
    elif sys.argv[1] == "ico":
        t = (1.0 + sqrt(5.0)) / 2.0
        points = [
            normalize(-1, t, 0),
            normalize(1, t, 0),
            normalize(-1, -t, 0),
            normalize(1, -t, 0),
            normalize(0, -1, t),
            normalize(0, 1, t),
            normalize(0, -1, -t),
            normalize(0, 1, -t),
            normalize(t, 0, -1),
            normalize(t, 0, 1),
            normalize(-t, 0, -1),
            normalize(-t, 0, 1),
        ]

        faces = [
            (points[0], points[11], points[5]),
            (points[0], points[5], points[1]),
            (points[0], points[1], points[7]),
            (points[0], points[7], points[10]),
            (points[0], points[10], points[11]),
            (points[1], points[5], points[9]),
            (points[5], points[11], points[4]),
            (points[11], points[10], points[2]),
            (points[10], points[7], points[6]),
            (points[7], points[1], points[8]),
            (points[3], points[9], points[4]),
            (points[3], points[4], points[2]),
            (points[3], points[2], points[6]),
            (points[3], points[6], points[8]),
            (points[3], points[8], points[9]),
            (points[4], points[9], points[5]),
            (points[2], points[4], points[11]),
            (points[6], points[2], points[10]),
            (points[8], points[6], points[7]),
            (points[9], points[8], points[1]),
        ]
    else:
        print("invalid type, must be oct or ico")
        exit(1)

    depth = int(sys.argv[2])
    if depth <= 0:
        print("depth must be larger than 0")
        exit(1)
    points = []
    for i in range(depth):
        tmpfaces = []
        for a, b, c in faces:
            d = get_middle_point(a, b)
            e = get_middle_point(b, c)
            f = get_middle_point(c, a)
            for A, B, C in [
                (a, d, f),
                (b, e, d),
                (c, f, e),
                (d, e, f)]:
                if i < (depth-1):
                    tmpfaces.append((A, B, C))
                else:
                    # last iteration
                    if sys.argv[3] in ["tri", "rat"]:
                        n1 = norm(cross((A[0]-B[0], A[1]-B[1], A[2]-B[2]), A))
                        n2 = norm(cross((B[0]-C[0], B[1]-C[1], B[2]-C[2]), B))
                        n3 = norm(cross((C[0]-A[0], C[1]-A[1], C[2]-A[2]), C))
                        alpha = pi - acos(n1[0]*n2[0]+n1[1]*n2[1]+n1[2]*n2[2])
                        beta = pi - acos(n2[0]*n3[0]+n2[1]*n3[1]+n2[2]*n3[2])
                        gamma = pi - acos(n3[0]*n1[0]+n3[1]*n1[1]+n3[2]*n1[2])
                        triangle_area = (alpha + beta + gamma - pi)
                    if sys.argv[3] in ["cc", "rat"]:
                        _, theta = circumcircle(A, B, C)
                        sphere_cap_area = 2 * pi * (1 - cos(theta))
                    if sys.argv[3] == "tri":
                        val = triangle_area
                    elif sys.argv[3] == "cc":
                        val = sphere_cap_area
                    elif sys.argv[3] == "rat":
                        val = triangle_area/sphere_cap_area
                    else:
                        print("invalid val, must be tri, cc or rat")
                        exit(1)
                    points.append((A, val))
                    points.append((B, val))
                    points.append((C, val))
                    tmpfaces.append((len(points)-3, len(points)-2, len(points)-1))
        faces = tmpfaces

    # the best possible ratio between a triangle and its circumcircle is:
    # 3/4*sqrt(3)/pi => 0.4135...
    if len(sys.argv) > 4:
        minval = float(sys.argv[4])
    else:
        minval = min([val for (_, val) in points])
    if len(sys.argv) > 5:
        maxval = float(sys.argv[5])
    else:
        maxval = max([val for (_, val) in points])

    print("ply")
    print("format ascii 1.0")
    print("comment minval: %f"%minval)
    print("comment maxval: %f"%maxval)
    print("element vertex %d" % len(points))
    print("property float x")
    print("property float y")
    print("property float z")
    print("property uchar red")
    print("property uchar green")
    print("property uchar blue")
    print("element face %d" % len(faces))
    print("property list uchar int vertex_index")
    print("end_header")
    for ((x, y, z), val) in points:
        r, g, b = jetmap(val, minval, maxval)
        print("%f %f %f %d %d %d" % (x, y, z, r, g, b))
    for a, b, c in faces:
        print("3 %d %d %d" % (a, b, c))


if __name__ == "__main__":
    main()
